/*
 *   IRC - Internet Relay Chat, modules/m_rehash.c
 *
 *   Copyright (C) 2000-2002 TR-IRCD Development
 *
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "s_conf.h"
#include "h.h"
#include "hook.h"
#include "resnew.h"
#include "throttle.h"

static int hookid_rehash_more = 0;

static struct Message _msgtab[] = {
    {MSG_REHASH, 0, MAXPARA, M_SLOW, 0L,
     m_unregistered, m_permission, m_rehash, m_ignore, m_ignore}
};

typedef struct {
    char *name;
    void (*function) ();
} RehashItem;

static void remove_szlines();
static void read_all_motd();
static void read_links();
static void do_rehash_everything();
static void do_rehash_help();
static void do_rehash_gc();

static RehashItem rlist[] = {
    {"DNS", restart_resolver},
    {"SZLINES", remove_szlines},
    {"TKLINES", flush_temp_klines},
    {"GC", do_rehash_gc},
    {"MOTD", read_all_motd},
    {"LINKS", read_links},
    {"AKILLS", do_rehash_akills},
    {"EXCLUDES", do_rehash_excludes},
    {"THROTTLES", throttle_rehash},
    {"ALL", do_rehash_everything},
    {"HELP", do_rehash_help},
    {NULL},
};

static int log_rehash = -1;

#ifndef STATIC_MODULES

char *_version = "$Revision: 5.13 $";

void _modinit(void)
{
    hookid_rehash_more = hook_add_event("doing rehash");
    mod_add_cmd(_msgtab);
    log_rehash = logevent_register("rehash", LOG_ALWAYS, LOG_LOGFILE | LOG_IRCLOG, LOG_NOTICE,                              
                                   "REHASH From %s\n");       
}

void _moddeinit(void)
{
    hook_del_event("doing rehash");
    mod_del_cmd(_msgtab);
    logevent_unregister(log_rehash);
}
#else
void m_rehash_init(void)
{
    hookid_rehash_more = hook_add_event("doing rehash");
    mod_add_cmd(_msgtab);
    log_rehash = logevent_register("rehash", LOG_ALWAYS, LOG_LOGFILE | LOG_IRCLOG, LOG_NOTICE,
                                   "REHASH From %s\n");  
}
#endif

static int help = 0;

static void remove_szlines()
{
    remove_szline("*", 1);
}

static void do_rehash_gc()
{
    block_garbage_collect(NULL);
}

static void read_all_motd()
{
    read_message_file(&(GeneralOpts.motd));
    read_message_file(&(GeneralOpts.shortmotd));
}

static void read_links()
{
    read_message_file(&(GeneralOpts.linksfile));
}

static void do_rehash_everything()
{
    restart_resolver();
    remove_szlines();
    flush_temp_klines();
    read_all_motd();
    read_links();
    do_rehash_akills();
    do_rehash_excludes();
    throttle_rehash();
}

static void do_rehash_help()
{
    help = 1;
}

static void rehash_more(aClient *sptr, char *more)
{
    struct hook_data thisdata;

    thisdata.source_p = sptr;
    thisdata.data = more;

    hook_call_event(hookid_rehash_more, &thisdata);
}

int m_rehash(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    help = 0;
    if (parc > 1) {
    int i, j;
	rehash_more(sptr, parv[1]);
	for (i = 0; rlist[i].name; i++) {
	    if (irc_strcmp(parv[1], rlist[i].name) == 0) {
		send_me_numeric(sptr, RPL_REHASHING, rlist[i].name);
		rlist[i].function();
		if (!help)
		    sendto_ops("%s is rehashing %s", parv[0], parv[1]);
	    }
	}
	if (help) {
	    send_me_notice(sptr, ":Rehash elements");
            for (j = 0; rlist[j].name; j++) 
                send_me_notice(sptr, ":%s", rlist[j].name);
        }
	return 0;
    } else {
	send_me_numeric(sptr, RPL_REHASHING, "Server config file");
	sendto_ops("%s is rehashing Server config file while whistling innocently", parv[0]);
	logevent_call(log_rehash, get_client_name(sptr, FALSE));
	return rehash(cptr, sptr, 0);
    }
    return 0;
}
