/************************************************************************
 *   IRC - Internet Relay Chat, newconf/config.c
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  (C) 1988 University of Oulu,Computing Center and Jarkko Oikarinen"
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/stat.h>
#include <limits.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "h.h"
#include "ircd_defs.h"
#include "tools.h"
#include "channel.h"
#include "confitem.h"
#include "class.h"
#include "event.h"
#include "hash.h"
#include "listener.h"
#include "modules.h"
#include "language.h"
#include "numeric.h"
#include "fd.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "fileio.h"
#include "y.tab.h"

FBFILE *conf_fbfile_in;

static void     clear_out_old_conf(void);

extern int scount;
extern int lineno;
extern char linebuf[];
extern int yyparse();		/* defined in y.tab.c */

/*
 * field breakup for ircd.conf file.
 */
static char *getfield(char *newline)
{
    static char *line = (char *) NULL;
    char *end, *field;

    if (newline)
	line = newline;

    if (line == (char *) NULL)
	return ((char *) NULL);

    field = line;
    if ((end = strchr(line, ':')) == NULL) {
	line = (char *) NULL;
	if ((end = strchr(field, '\n')) == (char *) NULL)
	    end = field + strlen(field);
    } else
	line = end + 1;
    *end = '\0';
    return (field);
}

/*
 * read_conf_old
 *
 * returns -1 on any error or else the fd opened from which to read the
 * configuration file from.  This may either be the file direct or one
 * end of a pipe from m4.
 */

#define MAXCONFLINKS 150

int read_conf_old(char *filename)
{

    static char quotes[9][2] = {
	{'b', '\b'}, {'f', '\f'}, {'n', '\n'},
	{'r', '\r'}, {'t', '\t'}, {'v', '\v'},
	{'\\', '\\'}, {0, 0}
    };

    char *tmp, *s;
    int i, dontadd;
    char line[512];
    aConfItem *aconf = NULL;
    FBFILE *file;

    file = fbopen(filename, "r");
    if (!file)
	return 0;

    while (fbgets(line, 512, file)) {
	if ((tmp = (char *) strchr(line, '\n')))
	    *tmp = '\0';
        if ((tmp = (char *) strchr(line, '\r')))
            *tmp = '\0';
	for (tmp = line; *tmp; tmp++) {
	    if (*tmp == '\\') {
		for (i = 0; quotes[i][0]; i++)
		    if (quotes[i][0] == *(tmp + 1)) {
			*tmp = quotes[i][1];
			break;
		    }
		if (!quotes[i][0])
		    *tmp = *(tmp + 1);
		if (!*(tmp + 1))
		    break;
		else
		    for (s = tmp; (*s = *(s + 1)); s++);
	    }
	}
	if (!*line || line[0] == '#' || line[0] == ' ')
	    continue;
	if (line[1] != ':') {
	    logevent_call(LogSys.badconfigline, line);
	    continue;
	}

	if (aconf)
	    free_conf(aconf);
	aconf = make_conf();
	tmp = getfield(line);
	if (!tmp)
	    continue;
	dontadd = 0;
	switch (*tmp) {
	    case 'G':		/* restricted gcos */
	    case 'g':
		aconf->status = CONF_GCOS;
		break;
	    case 'K':		/* the infamous klines */
	    case 'k':
		aconf->status = CONF_KILL;
		break;
	    case 'Q':		/* restricted nicknames */
	    case 'q':
		aconf->status = CONF_QUARANTINE;
		break;
	    case 'J':		/* jupitered channels */
	    case 'j':
		aconf->status = CONF_JUPECHAN;
		break;
	    case 'U':
	    case 'u':
		aconf->status = CONF_COMPAT_ULINE;
		break;
	    case 'H':
	    case 'h':
		aconf->status = CONF_COMPAT_HUB;
		break;
	    default:
		break;
	}
	if (IsIllegal(aconf))
	    continue;
	for (;;) {		/* Fake loop, that I can use break here --msa */
	    if ((tmp = getfield(NULL)) == NULL)
		break;
	    DupString(aconf->host, tmp);
	    if ((tmp = getfield(NULL)) == NULL)
		break;
	    DupString(aconf->passwd, tmp);
	    if ((tmp = getfield(NULL)) == NULL)
		break;
	    DupString(aconf->name, tmp);
	    if ((tmp = getfield(NULL)) == NULL)
		break;
	    break;
	    /*
	     * NOTREACHED
	     */
	}

	if ((aconf->status & CONF_KILL) && aconf->host) {
    char *host = host_field(aconf);
	    dontadd = 1;
	    switch (sortable(host)) {
		case 0:
		    l_addto_conf_list(&KList3, aconf, host_field);
		    break;
		case 1:
		    addto_conf_list(&KList1, aconf, host_field);
		    break;
		case -1:
		    addto_conf_list(&KList2, aconf, rev_host_field);
		    break;
	    }

	    MyFree(host);
	}
	if (aconf->name && (aconf->status & CONF_COMPAT_HUB)) {
    char *nm = name_field(aconf);
	    dontadd = 1;
	    add_host_to_oldconf_list(aconf->status, nm);
	    MyFree(nm);
	}
        if (aconf->host && (aconf->status & CONF_COMPAT_ULINE)) {
    char *host = host_field(aconf);
            dontadd = 1;
            add_host_to_oldconf_list(aconf->status, host);
            MyFree(host);
        }
	(void) collapse(aconf->host);
	(void) collapse(aconf->name);
	logevent_call(LogSys.add_confline, 
	       aconf->status, aconf->host, aconf->passwd, aconf->name, aconf->port, Class(aconf));
	if (!dontadd) {
	    aconf->next = GlobalConfItemList;
	    GlobalConfItemList = aconf;
	}
	aconf = NULL;
    }
    if (aconf)
	free_conf(aconf);
    fbclose(file);
    check_class();
    return 0;
}

static void set_default_conf(void)
{
    ServerOpts.ts_warn_delta = 60;
    ServerOpts.ts_max_delta = 300;
    ServerOpts.default_kline_time = 180;
    ServerOpts.motd_wait = 10;
    ServerOpts.motd_max = 3;
    ServerOpts.client_flood = 2560;
    ServerOpts.anti_nick_flood = 1;
    ServerOpts.max_nick_time = 20;
    ServerOpts.max_nick_changes = 5;
    ServerOpts.no_away_flood = 1;
    ServerOpts.max_away_time = 180;
    ServerOpts.max_away_count = 5;
    ServerOpts.no_oper_flood = 1;
    ServerOpts.save_maxclient_stats = 1;
    ServerOpts.save_maxclient_stats_time = 3600;
    ServerOpts.use_remoteconnect = 1;
    ServerOpts.use_registerfilter = 0;
    ServerOpts.dont_check_qline_remote = 1;
    ServerOpts.short_motd = 1;
    ServerOpts.hub = 0;
    ServerOpts.staffhide = 1;
    ServerOpts.wingate_notice = 1;
    ServerOpts.identd_use_tilde = 1;
    ServerOpts.identd_timeout = 30;
    ServerOpts.identd_complain = 0;
    ServerOpts.opers_can_choose_fakehost = 0;
    ServerOpts.throttle_time = 30;
    ServerOpts.throttle_count = 5;
    ServerOpts.anti_spambot = 0;
    ServerOpts.min_join_leave_time = 60;
    ServerOpts.max_join_leave_count = 25;
    ServerOpts.join_leave_count_expire_time = 120;
    ServerOpts.oper_spam_countdown = 0;
    ServerOpts.no_oper_jupiter = 1;

    ServicesConf.hide_ulined_servers = 1;

    ChannelConf.modhost = 0;
    ChannelConf.chanban = 0;
    ChannelConf.halfop = 1;
    ChannelConf.max_channels_per_user = 25;

    ServerHide.enable = 0;
    ServerHide.links_from_file = 0;
    ServerHide.flatten_links = 0;
    ServerHide.links_oper_only = 0;
    ServerHide.links_delay = 600;

    DupString(ChannelConf.default_quit_msg, DEF_QUITMSG);

    DupString(ServerOpts.server_kline_address, DEF_KLINEADDR);
    DupString(ServerOpts.network_kline_address, DEF_AKILLADDR);
    DupString(ServerOpts.staffhost, DEF_STAFFHOST);
    DupString(ServerOpts.monitor_host, DEF_MONHOST);
    DupString(ServerOpts.proxy_url, DEF_PROXYURL);

    ServerOpts.language = NULL;

    DupString(ServicesConf.operserv_other_host, DEF_OSHOST);
    DupString(ServicesConf.chanserv, DEF_CHANSERV);
    DupString(ServicesConf.nickserv, DEF_NICKSERV);
    DupString(ServicesConf.memoserv, DEF_MEMOSERV);
    DupString(ServicesConf.operserv, DEF_OPERSERV);
    DupString(ServicesConf.statserv, DEF_STATSERV);
    DupString(ServicesConf.services_name, DEF_SERVICES);

    DupString(AdminInfo.name, DEF_ADMINNAME);
    DupString(AdminInfo.email, DEF_ADMINMAIL);
    DupString(AdminInfo.description, DEF_ADMINDESC);

    memset(&ServerInfo.address, 0, sizeof(ServerInfo.address));
    ServerInfo.specific_ipv4_vhost = 0;
    memset(&ServerInfo.address6, 0, sizeof(ServerInfo.address6));
    ServerInfo.specific_ipv6_vhost = 0;

    memset(&ServerInfo.aliasname, 0, sizeof(ServerInfo.aliasname));

    return;
}

static void validate_conf(void)
{
    if (ServerInfo.name == NULL) {
	fprintf(stderr, "ERROR: No server name specified in serverinfo block.\n");
	exit(-1);
    }

    if (ServerInfo.description == NULL) {
	fprintf(stderr, "ERROR: No server description specified in serverinfo block.\n");
	exit(-1);
    }

    strlcpy_irc(me.info, ServerInfo.description, REALLEN);

    if (ServerInfo.networkname == NULL) {
	fprintf(stderr, "ERROR: No network name specified in serverinfo block.\n");
	exit(-1);
    }
#if 0
    if (ServerInfo.networkdesc == NULL) {
	fprintf(stderr, "ERROR: No network description specified in serverinfo block.\n");
	exit(-1);
    }
#endif
    if (ServerInfo.aliasname == NULL)
	DupString(ServerInfo.aliasname, "localhost");

    if (ServerInfo.servermode == 0)
	ServerInfo.servermode = SM_NATIVE;

    if (ServerOpts.language == NULL)
	me.lang = set_language(0);
    else
	me.lang = lang_parse(ServerOpts.language);

    return;
}

/*
 * read_conf_new() 
 *
 *
 * inputs       - file descriptor pointing to config file to use
 * output       - None
 * side effects - Read configuration file.
 */
static void read_conf_new(FBFILE * file)
{
    scount = lineno = 0;

    set_default_conf();		/* Set default values prior to conf parsing */
    yyparse();			/* Load the values from the conf */
    validate_conf();
    /* Some global values are also loaded here. */
    check_class();		/* Make sure classes are valid */
}

/*
 * yyerror
 *
 * inputs       - message from parser
 * output       - none
 * side effects - message to opers and log file entry is made
 */
void yyerror(char *msg)
{
    char newlinebuf[BUFSIZE];

    strip_tabs(newlinebuf, (const unsigned char *) linebuf, strlen(linebuf));

    sendto_lev(SNOTICE_LEV, "%d: %s on line: %s", lineno + 1, msg, newlinebuf);
    logevent_call(LogSys.yyerror, lineno + 1, msg, newlinebuf);
}

void read_conf_files(int cold)
{
    const char *filename;
    conf_fbfile_in = NULL;
    filename = GeneralOpts.configfile;
    if ((conf_fbfile_in = fbopen(filename, "r")) == NULL) {
	if (cold) {
	    logevent_call(LogSys.conferror, filename);
	    exit(-1);
	} else {
	    sendto_ops("Can't open file '%s' - aborting rehash!", filename);
	    return;
	}
    }
    if (!cold) {
	clear_out_old_conf();
        clear_oldconf_list();
    }

    read_conf_new(conf_fbfile_in);
    fbclose(conf_fbfile_in);
}

/*
 * clear_out_old_conf
 *
 * inputs       - none
 * output       - none
 * side effects - Clear out the old configuration
 */
static
void clear_out_old_conf(void)
{
    struct ConfItem **tmp = &GlobalConfItemList;
    struct ConfItem *tmp2;
    struct Class *cltmp;

    /*
     * We only need to free anything allocated by yyparse() here.
     * Reseting structs, etc, is taken care of by set_default_conf().
     */
    while ((tmp2 = *tmp)) {
	if (tmp2->clients) {
	    /*
	     ** Configuration entry is still in use by some
	     ** local clients, cannot delete it--mark it so
	     ** that it will be deleted when the last client
	     ** exits...
	     */
	    if (!(tmp2->status & CONF_CLIENT)) {
		*tmp = tmp2->next;
		tmp2->next = NULL;
	    } else
		tmp = &tmp2->next;
	    tmp2->status |= CONF_ILLEGAL;
	} else {
	    *tmp = tmp2->next;
	    free_conf(tmp2);
	}
    }

    /*
     * don't delete the class table, rather mark all entries
     * for deletion. The table is cleaned up by check_class. - avalon
     */
    for (cltmp = classes->next; cltmp; cltmp = cltmp->next)
	MaxLinks(cltmp) = -1;

    /* operator{} and class{} blocks are freed above */
    /* clean out listeners */

    close_listeners();
}

int conf_yy_fatal_error(char *msg)
{
#ifdef IMPOSSIBLE_CASES

    /*
     * This message does not mean anything, since the config file is
     * obviously successfully read.
     * h7 has set this already to return 0, I just put it between defines.
     * -TimeMr14C, 26/07/2002
     */

    sendto_lev(DEBUG_LEV, "lexer barfed. lets leave it at that for now");
    if (msg)
	sendto_lev(DEBUG_LEV, "Lexer error: %s", msg);
#endif
    return 0;
}
