/************************************************************************
 *   IRC - Internet Relay Chat, protocol/pelennor.c
 *
 *   Copyright (C) 2000-2002 TR-IRCD Development
 *
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "h.h"
#include "comply.h"
#include "s_conf.h"
#include "hook.h"

static int log_greet = -1;        
static int log_remove = -1;                 
static int log_skip = -1;     
static int log_fail = -1;
static int log_delete = -1;
static int log_unload = -1;
static int log_ignore = -1;
static int log_untok = -1;
static int log_retok = -1;             
static int log_unuse = -1;

static int do_introduce_client(struct hook_data *thisdata);
static int do_sendnick_TS(struct hook_data *thisdata);
static int do_inform_remote_servers(struct hook_data *thisdata);
static int do_inform_other_servers(struct hook_data *thisdata);
static int do_continue_server_estab(struct hook_data *thisdata);
static int do_start_server_estab(struct hook_data *thisdata);
static int do_connect_to_server(struct hook_data *thisdata);

#define MSG_CHATOPS 	"CHATOPS"
#define MSG_HTM  	"HTM"

#define TOK1_CHATOPS    "C"	/* Please note that this will collide with JUPITER */

#ifndef HAVE_SHL_LOAD
static char *unloadtab[] = {
    "m_svsjoin.so",
    "m_operdo.so",
    "m_jupiter.so",
    "m_unjupiter.so",
    "m_exclude.so",
    "m_rexclude.so",
    "m_rping.so",
    "m_rnotice.so",
    "m_admins.so",
    "m_sadmins.so",
    NULL
};
#else
static char *unloadtab[] = {
    "m_svsjoin.sl",
    "m_operdo.sl",
    "m_jupiter.sl",
    "m_unjupiter.sl",
    "m_exclude.sl",
    "m_rexclude.sl",
    "m_rping.sl",
    "m_rnotice.sl",
    "m_admins.sl",
    "m_sadmins.sl",
    NULL
};
#endif

#ifdef STATIC_MODULES
static char *unloadmsgtab[] = { 
	MSG_RPING, MSG_RNOTICE, MSG_ADMINS, MSG_SADMINS, 
	MSG_SVSJOIN, MSG_OPERDO, MSG_JUPITER, MSG_UNJUPE, 
	MSG_EXCLUDE, MSG_REXCLUDE, NULL };
static char *unloadtoktab[] = { 
	TOK1_RPING, TOK1_RNOTICE, TOK1_ADMINS, TOK1_SADMINS, 
	TOK1_SVSJOIN, NULL, TOK1_JUPITER, TOK1_UNJUPE, 
	TOK1_EXCLUDE, TOK1_REXCLUDE, NULL };
#endif

static int m_chatops(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    char *message = parc > 1 ? parv[1] : NULL;

    if (BadPtr(message)) {
	if (MyClient(sptr))
	    send_me_numeric(sptr, ERR_NEEDMOREPARAMS, MSG_CHATOPS);
	return 0;
    }

    if (strlen(message) > TOPICLEN)
	message[TOPICLEN] = '\0';

    sendto_serv_butone(cptr, sptr, TOK1_CHATOPS, ":%s", message);
    sendto_operators(0, "ChatOps", "from %C: %s", sptr, message);
    return 0;
}

static struct Message chatops_tab[] = {
    {MSG_CHATOPS, 0, MAXPARA, M_SLOW | M_IDLE_RESET, 0L,
     m_ignore, m_ignore, m_ignore, m_chatops, m_ignore}
};

static struct Message htm_tab[] = {
    {MSG_HTM, 0, 1, M_SLOW | M_IDLE_RESET, 0L,
     m_ignore, m_ignore, m_ignore, m_ignore, m_ignore}
};

static int do_stop_module(struct hook_data *thisdata)
{
    char *name = thisdata->name;
    int i;

    for (i = 0; unloadtab[i]; i++)
	if (!irc_strcmp(name, unloadtab[i]))
	    return 1;
    return 0;
}

void proto_modinit()
{
    if (!GeneralOpts.protocol_in_use) {
        hook_add_hook("continue server estab", (hookfn *) do_continue_server_estab);
        hook_add_hook("start server estab", (hookfn *) do_start_server_estab);
        hook_add_hook("sendnick TS", (hookfn *) do_sendnick_TS);
        hook_add_hook("introduce client", (hookfn *) do_introduce_client);
        hook_add_hook("inform remote servers", (hookfn *) do_inform_remote_servers);
        hook_add_hook("inform other servers", (hookfn *) do_inform_other_servers);
        hook_add_hook("connect to server", (hookfn *) do_connect_to_server);
        hook_add_hook("load module", (hookfn *) do_stop_module);

        log_greet = logevent_register("protocol init", LOG_ON_LOG, LOG_IRCLOG|LOG_STDOUT, LOG_INFO,
                        "Initialized Bahamut Pelennor compatibility TR-IRCD 5 Module");
        log_remove = logevent_register("protocol modremove", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Module %s is being removed");
        log_skip = logevent_register("protocol modskip", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Skipped non-loaded module %s");
        log_fail = logevent_register("protocol modfail", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Failed in unloading %s");
        log_delete = logevent_register("protocol delmsg", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Message %s being deleted");
        log_unload = logevent_register("protocol modunload", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Unsupported modules are now removed");
        log_ignore = logevent_register("protocol ignore", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Added compatibility commands: %s");     
        log_untok = logevent_register("protocol untokenize", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "%s");
        log_retok = logevent_register("protocol retokenize", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Retokenized token %s, added %s");
        log_unuse = logevent_register("protocol unuse", LOG_ON_LOG, LOG_IRCLOG, LOG_INFO,
                        "Removed incompatible commands: %s");
	logevent_call(log_greet);
    }
    ServerInfo.servermode = SM_PELENNOR;
    return;
}

int _comply_unload_modules()
{
int i = 0;
#ifndef STATIC_MODULES
int j = 0;
    for (i = 0; unloadtab[i]; i++) {
        j = unload_one_module(unloadtab[i], 0);
        if (j == 0)
            logevent_call(log_remove, unloadtab[i]);
        else if (j == -1)
            logevent_call(log_skip, unloadtab[i]);
        else if (j == -2) {
            logevent_call(log_fail, unloadtab[i]);
            return 0;
        }
    }
#else
char *token = NULL;
    for (i = 0; unloadmsgtab[i]; i++) {
        token = unloadtoktab[i];
        mod_del_msg(unloadmsgtab[i]);
        tok1_msgtab[(u_char) *token].msg = NULL;
        logevent_call(log_delete, unloadmsgtab[i]);
    }
    logevent_call(log_unload);
#endif
    return 1;
}

int _comply_ignore_messages()
{

    mod_add_cmd(chatops_tab);
    mod_add_cmd(htm_tab);
    logevent_call(log_ignore, "CHATOPS HTM");
    return 1;
}

int _comply_untokenize()
{
    logevent_call(log_untok, "No commands are untokenized. Pelennor does not support tokens");
    return 1;
}

int _comply_retokenize()
{

    tok1_msgtab[(u_char) TOK1_CHATOPS].cmd = MSG_CHATOPS;
    tok1_msgtab[(u_char) TOK1_CHATOPS].msg = chatops_tab;
    logevent_call(log_retok, "C", "CHATOPS");
    return 1;
}

int _comply_unuse_messages()
{
    char *token;

    token = TOK1_NETSET;
    mod_del_msg(MSG_NETSET);
    tok1_msgtab[(u_char) *token].msg = NULL;

    token = TOK1_CLIENT;
    mod_del_msg(MSG_CLIENT);
    tok1_msgtab[(u_char) *token].msg = NULL;

    token = TOK1_GLINE;
    mod_del_msg(MSG_GLINE);
    tok1_msgtab[(u_char) *token].msg = NULL;

    token = TOK1_UNGLINE;
    mod_del_msg(MSG_UNGLINE);
    tok1_msgtab[(u_char) *token].msg = NULL;

    token = TOK1_TMODE;
    mod_del_msg(MSG_TMODE);
    tok1_msgtab[(u_char) *token].msg = NULL;

    logevent_call(log_unuse, "GLINE UNGLINE CLIENT NETSET TMODE");
    return 1;
}

int _comply_rehash()
{
    ChannelConf.owners = 0;
    ChannelConf.protected = 0;
    ChannelConf.excepts = 0;
    ChannelConf.invites = 0;
    ChannelConf.halfop = 0;
    ChannelConf.chanban = 0;
    ChannelConf.modhost = 0;
    ChannelConf.floodmode = 0;

    if (!GeneralOpts.protocol_in_use) {
        logevent_unregister(log_greet);
        logevent_unregister(log_remove);
        logevent_unregister(log_skip);
        logevent_unregister(log_fail);
        logevent_unregister(log_delete);
        logevent_unregister(log_unload);
        logevent_unregister(log_ignore);
        logevent_unregister(log_untok); 
        logevent_unregister(log_retok);
        logevent_unregister(log_unuse);
    }

    return 1;
}

static int do_introduce_client(struct hook_data *thisdata)
{
    struct Client *cptr = thisdata->client_p;
    struct Client *sptr = thisdata->source_p;
    struct User *user = thisdata->user;
    char *nick = thisdata->name;
    struct Client *nsptr;
    char ubuf[32];
    unsigned long li = 0;

    send_umode(NULL, sptr, 0, SEND_UMODES, ubuf);

    if (!*ubuf) {
        ubuf[0] = '+';
        ubuf[1] = '\0';
    }
    hash_check_watch(sptr, RPL_LOGON);
    if (!IsClientIpv6(sptr))
        li = htonl(IN4_ADDR(sptr->ip));

#ifdef USE_FAKEHOST
    sendto_serv_butone(cptr, NULL, TOK1_NICK,
                       "%s %d %T %s %s %s %s %lu %lu :%s",
                       nick, sptr->hopcount + 1, sptr, ubuf,
                       user->username, user->fakehost,
                       user->server, sptr->user->servicestamp, li, sptr->info);
    sendto_service(SERVICE_SEE_RNICKS, 0, NULL, NULL, TOK1_NICK,
                   "%s %d %T %s %s %s %s %s %lu :%s", nick,
                   sptr->hopcount + 1, sptr, ubuf,
                   user->username, user->host, user->fakehost,
                   user->server, sptr->user->servicestamp, sptr->info);
#else
    sendto_serv_butone(cptr, NULL, TOK1_NICK,
                       "%s %d %T %s %s %s %s %lu %lu :%s", nick,
                       sptr->hopcount + 1, sptr, ubuf, user->username,
                       user->host, user->server, 
		       sptr->user->servicestamp, li, sptr->info);
    sendto_service(SERVICE_SEE_RNICKS, 0, NULL, NULL, TOK1_NICK,
                   "%s %d %T %s %s %s %s %lu :%s", nick, sptr->hopcount + 1,
                   sptr, ubuf, user->username, user->host,
                   user->server, sptr->user->servicestamp, sptr->info);
#endif

    /* FIXME: This can make a user identify to any NickServ -TimeMr14C */

    if (MyClient(sptr)) {
        if (sptr->nspass[0]
            && (nsptr = find_person(ServicesConf.nickserv))) {
            sendto_one_server(nsptr, sptr, TOK1_PRIVMSG, "%s@%s :SIDENTIFY %s",
                              ServicesConf.nickserv,
                              ServicesConf.services_name, sptr->nspass);
        }

        memset(sptr->passwd, '\0', PASSWDLEN);

    }

    if (MyClient(cptr) && ubuf[1])
        send_umode(cptr, sptr, 0, SEND_UMODES, ubuf);
    return 0;
}

static int do_sendnick_TS(struct hook_data *thisdata)
{
    struct Client *cptr = thisdata->client_p;
    struct Client *acptr = thisdata->source_p;
    char ubuf[32];
    unsigned long li = 0;

    if (IsPerson(acptr)) {
        send_umode(NULL, acptr, 0, SEND_UMODES, ubuf);
        if (!*ubuf) {           /* trivial optimization - Dianora */
            ubuf[0] = '+';
            ubuf[1] = '\0';
        }
        if (!IsClientIpv6(acptr))
            li = htonl(IN4_ADDR(acptr->ip));
#ifdef USE_FAKEHOST
        sendto_one_server(cptr, NULL, TOK1_NICK,
                          "%s %d %T %s %s %s %s %lu %lu :%s",
                          acptr->name, acptr->hopcount + 1, acptr, ubuf,
                          acptr->user->username, acptr->user->fakehost, acptr->user->server,
                          acptr->user->servicestamp, li, acptr->info);
#else
        sendto_one_server(cptr, NULL, TOK1_NICK,
                          "%s %d %T %s %s %s %s %lu %lu :%s",
                          acptr->name, acptr->hopcount + 1, acptr, ubuf,
                          acptr->user->username, acptr->user->host, acptr->user->server,
                          acptr->user->servicestamp, li, acptr->info);
#endif
    }
    return 0;
}

static int do_inform_remote_servers(struct hook_data *thisdata)
{
    struct Client *bclient_p;
    dlink_node *ptr;
    struct Client *client_p = thisdata->client_p;
    struct Client *target_p = thisdata->aclient_p;
    char *name = thisdata->name;
    struct ConfItem *aconf;

    for (ptr = serv_list.head; ptr; ptr = ptr->next) {
        bclient_p = ptr->data;

        if (bclient_p == client_p)
            continue;
        if (!(aconf = bclient_p->serv->nline)) {
            sendto_gnotice("Lost N-line for %s on %s. Closing",
                           get_client_name(client_p, HIDE_IP), name);
            return exit_client(client_p, client_p, "Lost N line");
        }
        sendto_one_server(bclient_p, target_p->servptr, TOK1_SERVER, "%C %d :%s",
                          target_p, target_p->hopcount + 1, target_p->info);

    }
    return 0;
}

static int do_inform_other_servers(struct hook_data *thisdata)
{
    struct Client *cptr = thisdata->client_p;

    sendto_flag_serv_butone(cptr, 0, 0, &me, TOK1_SERVER,
                                "%C 2 :%s", cptr, cptr->info);
    return 0;
}

static int do_continue_server_estab(struct hook_data *thisdata)
{
    struct Client *client_p = thisdata->client_p;
    struct Client *target_p;

    for (target_p = &me; target_p; target_p = target_p->prev) {
        /* target_p->from == target_p for target_p == client_p */
        if (target_p->from == client_p)
            continue;
        if (IsServer(target_p)) {
            sendto_one_server(client_p, target_p->servptr,
                              TOK1_SERVER, "%C %d :%s",
                              target_p, target_p->hopcount + 1, target_p->info);
        }
    }
    return 0;
}

static int do_start_server_estab(struct hook_data *thisdata)
{
    struct Client *client_p = thisdata->client_p;
    struct ConfItem *aconf = thisdata->confitem;
    int emptypass = thisdata->check;
    int dontwantflags = 0;

    if (!emptypass)
        sendto_one(client_p, "%s %s :TS", MSG_PASS, aconf->spasswd);

#ifdef HAVE_ENCRYPTION_ON
    dontwantflags |= (IsConfEncrypted(aconf) ? 0 : CAPAB_DKEY);
#endif

    dontwantflags |= CAPAB_TRIRCD5;
    dontwantflags |= CAPAB_TRIRCD4;

    send_capab_to(client_p, dontwantflags);
    sendto_one_server(client_p, NULL, TOK1_SERVER, "%C 1 :%s", &me, me.info);
    sendto_one_server(client_p, NULL, TOK1_SVINFO, "%d %d 0 :%lu",
                      TS_CURRENT, TS_MIN, (unsigned long) timeofday);
    return 0;
}

static int do_connect_to_server(struct hook_data *thisdata)
{
#ifdef HAVE_ENCRYPTION_ON
    struct Client *client_p = thisdata->client_p;
#endif

    do_start_server_estab(thisdata);

#ifdef HAVE_ENCRYPTION_ON
    if (!CanDoDKEY(client_p) || !WantDKEY(client_p));
    else {
        SetNegoServer(client_p);        /* VERY IMPORTANT THAT THIS IS HERE */
        sendto_one(client_p, "DKEY START");
    }
#endif
    return 0;
}
