/************************************************************************
 *   IRC - Internet Relay Chat, server/s_user.c
 *
 *   Copyright (C) 2000-2002 TR-IRCD Development
 *
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers. 
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  $Id: s_user.c,v 5.24 2003/05/04 23:06:40 tr-ircd Exp $
 */

#include <assert.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "tools.h"
#include "h.h"
#include "channel.h"
#include "class.h"
#include "fd.h"
#include "hash.h"
#include "listener.h"
#include "msg.h"
#include "numeric.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "throttle.h"
#include "supported.h"
#include "md5.h"
#include "packet.h"
#include "hook.h"
#include "language.h"

#define IRCD_BUFSIZE 512

static int hookid_introduce_client = 0;

static char buf[BUFSIZE];

static void user_welcome(struct Client *source_p, struct ConfItem *aconf);
static void manage_faked_oper_hostname(struct Client *sptr, struct ConfItem *pwaconf);

static int valid_hostname(const char *hostname);
static int valid_username(const char *username);
static int check_drone_PB(char *username, char *gcos);
static int introduce_client(struct Client *client_p, struct Client *source_p,
			    struct User *user, char *nick);

int user_modes[] = {
    UMODE_o, 'o',		/* ops, realops */
    UMODE_i, 'i',		/* invisible on who/names */
    UMODE_w, 'w',		/* wallops */
    UMODE_r, 'r',		/* registered and identified nickname */
    UMODE_R, 'R',		/* listens to registered and identified
				 * nicknames */
    UMODE_a, 'a',		/* nick is services admin */
    UMODE_A, 'A',		/* nick is server admin */
    UMODE_n, 'n',
    UMODE_h, 'h',		/* helper */
#ifdef USE_FAKEHOST
    UMODE_x, 'x',		/* nick has hidden hostname */
#endif
#ifdef USE_GREEKLISH
    UMODE_t, 't',		/* Users wants to convert text to greeklish */
#endif
    0, 0
};

void init_user(void)
{
    hookid_introduce_client = hook_add_event("introduce client");
    return;
}

/*
 * ** register_local_user
 * **      This function is called when both NICK and USER messages
 * **      have been accepted for the client, in whatever order. Only
 * **      after this, is the USER message propagated.
 * **
 * **      NICK's must be propagated at once when received, although
 * **      it would be better to delay them too until full info is
 * **      available. Doing it is not so simple though, would have
 * **      to implement the following:
 * **
 * **      (actually it has been implemented already for a while) -orabidoo
 * **
 * **      1) user telnets in and gives only "NICK foobar" and waits
 * **      2) another user far away logs in normally with the nick
 * **         "foobar" (quite legal, as this server didn't propagate
 * **         it).
 * **      3) now this server gets nick "foobar" from outside, but
 * **         has already the same defined locally. Current server
 * **         would just issue "KILL foobar" to clean out dups. But,
 * **         this is not fair. It should actually request another
 * **         nick from local user or kill him/her...
 */

int register_local_user(struct Client *client_p, struct Client *source_p,
			char *nick, char *username)
{
    struct ConfItem *aconf;
    struct User *user = source_p->user;
    char *p;
    char passarr[PASSWDLEN];
    char tmpstr2[IRCD_BUFSIZE];
    char ipaddr[HOSTIPLEN];
    int status;
    dlink_node *ptr;
    dlink_node *m;
#ifdef USE_FAKEHOST
    char newhostarr[HOSTLEN];
    char *q;
#endif

    assert(0 != source_p);
    assert(0 != source_p);
    assert(source_p->username != username);

    user->last = timeofday;

    /* pointed out by Mortiis, never be too careful */
    if (strlen(username) > USERLEN)
	username[USERLEN] = '\0';

    if (!valid_hostname(source_p->sockhost)) {
	send_me_notice(source_p, ":*** Notice -- You have an illegal character in your hostname");

	strlcpy_irc(source_p->sockhost, source_p->hostip, HOSTIPLEN);
    }

    if (strchr(source_p->hostip, ':'))
	SetClientIpv6(source_p);

    strlcpy_irc(user->host, source_p->sockhost, HOSTLEN);

#ifdef USE_FAKEHOST
    q = user->host;
    strlcpy_irc(user->fakehost, calcmask(q, newhostarr), HOSTLEN);
#endif

    if ((status = check_client(client_p, source_p, username)) < 0)
	return (CLIENT_EXITED);


    ptr = source_p->confs.head;
    aconf = ptr->data;

    source_p->allow_read = MAX_FLOOD_PER_SEC_I;

    if (aconf == NULL) {
	exit_client(source_p, &me, "*** Not Authorized");
	return (CLIENT_EXITED);
    }

    if (!IsGotId(source_p)) {
	if (!ServerOpts.identd_use_tilde && ServerOpts.identd_complain) {
	    ircstp->is_ref++;
	    send_me_notice(source_p,
			   ":*** Notice -- You need to install identd to use this server");
	    exit_client(source_p, &me, "Install identd");
	    return (CLIENT_EXITED);
	} else if (!ServerOpts.identd_use_tilde && !ServerOpts.identd_complain) {
	    strlcpy_irc(source_p->username, username, USERLEN);
	} else {
	    *(source_p->username) = '~';
	    strlcpy_irc(&source_p->username[1], username, USERLEN - 1);
	}
	source_p->username[USERLEN] = '\0';
    }

    /* password check */
    if (!BadPtr(aconf->passwd) && (p = source_p->passwd) &&
	strcmp(calcpass(p, passarr), aconf->passwd)) {
	ircstp->is_ref++;
	send_me_numeric(source_p, ERR_PASSWDMISMATCH);
	exit_client(source_p, &me, "Bad Password");
	return CLIENT_EXITED;
    }
    memset(source_p->passwd, 0, sizeof(source_p->passwd));

    /* Limit clients */
    /*
     * We want to be able to have servers and F-line clients
     * connect, so save room for "buffer" connections.
     * Smaller servers may want to decrease this, and it should
     * probably be just a percentage of the MAXCLIENTS...
     *   -Taner
     */
    /* Except "F:" clients */
    if ((((Count.local + 1) >= GeneralOpts.maxclients + MAX_BUFFER)
	 || ((Count.local + 1) >= (GeneralOpts.maxclients - 5))) && !IsConfExemptLimits(aconf)) {
	sendto_lev(REJ_LEV, "Too many clients, rejecting %s[%s].", nick, source_p->sockhost);

	ircstp->is_ref++;
	exit_client(source_p, &me, "Sorry, server is full - try later");
	return (CLIENT_EXITED);
    }

    /* valid user name check */

    if (!valid_username(source_p->username)) {
	sendto_lev(REJ_LEV, "Invalid username: %s (%s@%s)",
		   nick, source_p->username, source_p->sockhost);
	ircstp->is_ref++;
	ircsprintf(tmpstr2, "Invalid username [%s]", source_p->username);
	exit_client(source_p, &me, tmpstr2);
	return (CLIENT_EXITED);
    }

    if (check_drone_PB(source_p->username, source_p->info)) {
	sendto_lev(REJ_LEV, "Rejecting acebot-style drone: %^C [%s]", source_p, source_p->info);
	exit_client(source_p, &me, "You match the pattern of a known"
		    " trojan, please check your system.");
	return (CLIENT_EXITED);
    }

    SetClient(source_p);

    strlcpy_irc(user->username, source_p->username, USERLEN);

    /* end of valid user name check */

    m = dlinkFind(&unknown_list, source_p);
    if (m != NULL) {
        Count.unknown--;
        dlinkDelete(m, &unknown_list);
        dlinkAdd(source_p, m, &lclient_list);
    }

    inetntop(source_p->aftype, &IN_ADDR(source_p->ip), ipaddr, HOSTIPLEN);
    sendto_lev(CCONN_LEV, "Client connecting: %^C [%s] {%d} [%s]",
	       source_p, ipaddr, get_client_class(source_p), source_p->info);

    if ((++Count.local) > Count.max_loc) {
	Count.max_loc = Count.local;
	if (!(Count.max_loc % 10))
	    sendto_lev(DEBUG_LEV, "New Max Local Clients: %d", Count.max_loc);
    }

    source_p->pingval = get_client_ping(source_p);
    source_p->sendqlen = get_sendq(source_p);

    /* XXX source_p->servptr is &me, since local client */
    source_p->servptr = find_server(user->server);
    add_client_to_llist(&(source_p->servptr->serv->users), source_p);

    source_p->servptr->serv->usercnt++;

    /* Increment our total user count here */
    if (++Count.total > Count.max_tot)
	Count.max_tot = Count.total;

    assign_localuser_identity(source_p);
    SetHasID(source_p);

    if (source_p->langstring[0]) {
	source_p->lang = lang_parse(source_p->langstring);
	if (source_p->lang)
	    sendto_one_person(source_p, NULL, TOK1_NOTICE, "AUTH :*** Setting language to %s",
				  source_p->langstring);
    }

    user_welcome(source_p, aconf);

    return (introduce_client(client_p, source_p, user, nick));
}

/*
 * register_remote_user
 *
 * inputs
 * output
 * side effects - This function is called when a remote client
 *                is introduced by a server.
 */
int register_remote_user(struct Client *client_p, struct Client *source_p,
			 char *nick, char *username)
{
    struct User *user = source_p->user;
    struct Client *target_p;

    assert(0 != source_p);
    assert(source_p->username != username);

    user->last = timeofday;

    /* pointed out by Mortiis, never be too careful */
    if (strlen(username) > USERLEN)
	username[USERLEN] = '\0';

    strlcpy_irc(source_p->username, username, USERLEN);
    strlcpy_irc(user->username, username, USERLEN);

    SetClient(source_p);

    /* Increment our total user count here */
    if (++Count.total > Count.max_tot)
	Count.max_tot = Count.total;

    source_p->servptr = find_server(user->server); 

    if (source_p->servptr == NULL) {
	sendto_ops("Ghost killed: %s on invalid server %s", source_p->name, source_p->user->server);

	kill_client(client_p, source_p, "%s (Server doesn't exist)", me.name);

	source_p->flags |= FLAGS_KILLED;
	return exit_client(source_p, &me, "Ghost");
    }

    if ((target_p = find_server(user->server)) && target_p->from != source_p->from) {
	sendto_lev(DEBUG_LEV, "Bad User [%s] :%s USER %s@%s %s, != %s[%s]",
		   client_p->name, nick, source_p->username,
		   source_p->sockhost, user->server, target_p->name, target_p->from->name);
	kill_client(client_p, source_p,
		    ":%s (%s != %s[%s] USER from wrong direction)", me.name,
		    user->server, target_p->from->name, target_p->from->sockhost);

	source_p->flags |= FLAGS_KILLED;
	return exit_client(source_p, &me, "USER server wrong direction");

    }
    /*
     * Super GhostDetect:
     * If we can't find the server the user is supposed to be on,
     * then simply blow the user away.        -Taner
     */
    if (!target_p) {
	kill_client(client_p, source_p, "%s GHOST (no server found)", me.name, user->server);
	sendto_ops("No server %s for user %s[%s@%s] from %s",
		   user->server, source_p->name, source_p->username,
		   source_p->sockhost, source_p->from->name);
	source_p->flags |= FLAGS_KILLED;
	return exit_client(source_p, &me, "Ghosted Client");
    }

    add_client_to_llist(&(source_p->servptr->serv->users), source_p);

    source_p->servptr->serv->usercnt++;

    if (ServerOpts.use_remoteconnect) {
        sendto_lev(CCONN_LEV, "Client connecting: %^C [%s] {0} [%s]",
                source_p,
                (IsClientIpv6(source_p) ? source_p->hostip : inetntoa((char *) &IN_ADDR(source_p->ip))),
                source_p->info);
    }

    return (introduce_client(client_p, source_p, user, nick));
}

/*
 * introduce_client
 *
 * inputs       -
 * output       -
 * side effects - This common function introduces a client to the rest
 *                of the net, either from a local client connect or
 *                from a remote connect.
 */
static int introduce_client(struct Client *cptr, struct Client *sptr, struct User *user, char *nick)
{
    struct hook_data thisdata;
    thisdata.client_p = cptr;
    thisdata.source_p = sptr;
    thisdata.user = user;
    thisdata.name = nick;

    if ((NOW - Count.day) > 86400) {
	Count.today = 0;
	Count.day = NOW;
    }
    if ((NOW - Count.week) > 604800) {
	Count.weekly = 0;
	Count.week = NOW;
    }
    if ((NOW - Count.month) > 2592000) {
	Count.monthly = 0;
	Count.month = NOW;
    }
    if ((NOW - Count.year) > 31536000) {
	Count.yearly = 0;
	Count.year = NOW;
    }
    Count.today++;
    Count.weekly++;
    Count.monthly++;
    Count.yearly++;

    return hook_call_event(hookid_introduce_client, &thisdata);

}

/* 
 * valid_hostname - check hostname for validity
 *
 * Inputs       - pointer to user
 * Output       - YES if valid, NO if not
 * Side effects - NONE
 *
 * NOTE: this doesn't allow a hostname to begin with a dot and
 * will not allow more dots than chars.
 */
static int valid_hostname(const char *hostname)
{
    const char *p = hostname;

    assert(0 != p);

    if ('.' == *p)
	return 0;

    while (*p) {
	if (!IsHostChar(*p))
	    return 0;
	p++;
    }

    return 1;
}

/* 
 * valid_username - check username for validity
 *
 * Inputs       - pointer to user
 * Output       - YES if valid, NO if not
 * Side effects - NONE
 * 
 * Absolutely always reject any '*' '!' '?' '@' in an user name
 * reject any odd control characters names.
 * Allow '.' in username to allow for "first.last"
 * style of username
 */
static int valid_username(const char *username)
{
    int dots = 0;
    const char *p = username;

    assert(0 != p);

    if ('~' == *p)
	++p;

    /* reject usernames that don't start with an alphanum
     * i.e. reject jokers who have '-@somehost' or '.@somehost'
     * or "-hi-@somehost", "h-----@somehost" would still be accepted.
     */
    if (!IsAlNum(*p))
	return 0;

    while (*++p) {
	if ((*p == '.')) {
	    dots++;
	    if (dots > 1)
		return 0;
	    if (!IsUserChar(p[1]))
		return 0;
	} else if (!IsUserChar(*p))
	    return 0;
    }
    return 1;
}

/*
 * Returns 1 if the user matches a drone style
 * discovered by PB@DAL.net
 */
static int check_drone_PB(char *username, char *gcos)
{
    unsigned char *x;

    if (*username == '~')
	username++;

    if (strlen(username) <= 2)
	return 0;

    /* verify that it's all uppercase leters */
    for (x = (unsigned char *) username; *x; x++) {
	if (*x < 'A' || *x > 'Z')
	    return 0;
    }

    if (strcmp(username, gcos))
	return 0;

    return 1;
}

/*
 * do_local_user
 *
 * inputs       -
 * output       -
 * side effects -
 */
int do_local_user(char *nick, aClient *cptr, aClient *sptr, char *username, char *host,
#ifdef USE_FAKEHOST
		  char *mask,
#endif
		  char *server, unsigned long serviceid, char *realname)
{
    struct User *user;

    user = make_user(sptr);

    if (!IsUnknown(sptr)) {
	send_me_numeric(sptr, ERR_ALREADYREGISTRED);
	return 0;
    }
    sptr->umode |= UMODE_i;	/* invisible IS default. always */
#ifdef USE_FAKEHOST
    sptr->umode |= UMODE_x;
#endif
    if (ServerOpts.use_registerfilter)
	sptr->umode |= UMODE_R;

    strlcpy_irc(user->host, host, sizeof(user->host));
    user->server = me.name;

    server[strlen(server) - 1] = '\0';

    if ((ServerInfo.aliasname[0] != '\0') && (server[0] != '\0')) {
	if (!irc_strcmp(server + 1, ServerInfo.aliasname))
	    sptr->protoflags |= PFLAGS_ALIASED;
    }

    Count.invisi++;

#ifdef USE_FAKEHOST
    if (mask) 
	strlcpy_irc(user->fakehost, mask, sizeof(user->fakehost));
#endif

    user->server = me.name;

    strlcpy_irc(sptr->info, realname, REALLEN);

    sptr->user->servicestamp = serviceid;
    sptr->oflag = 0;

    if (sptr->name[0]) {
	strlcpy_irc(user->username, username, USERLEN);
	/* NICK already received, now I have USER... */
	return register_local_user(cptr, sptr, sptr->name, username);
    } else if (!IsGotId(sptr)) {
	strlcpy_irc(user->username, username, USERLEN);
	strlcpy_irc(sptr->username, username, USERLEN);
	return 0;
    } else
	return 0;
}

/*
 * do_remote_user
 *
 * inputs       -
 * output       -
 * side effects -
 */

int do_remote_user(char *nick, aClient *client_p, aClient *source_p, char *username, char *host,
#ifdef USE_FAKEHOST
		   char *mask,
#endif
		   char *server, unsigned long serviceid, char *realname)
{
    struct User *user;

    assert(0 != source_p);
    assert(source_p->username != username);

    user = make_user(source_p);

    /*
     * coming from another server, take the servers word for it
     */
    user->server = find_or_add(server);

    strlcpy_irc(source_p->sockhost, host, HOSTLEN);
    strlcpy_irc(source_p->info, realname, REALLEN);
    strlcpy_irc(user->host, host, HOSTLEN);

#ifdef USE_FAKEHOST
    if (mask)
	strlcpy_irc(user->fakehost, mask, HOSTLEN);
#endif

    source_p->user->servicestamp = serviceid;

    return register_remote_user(client_p, source_p, source_p->name, username);
}

static void manage_faked_oper_hostname(aClient *sptr, aConfItem *pwaconf)
{
    if (!ServerOpts.staffhide)
	return;

    if (!IsConfDoSpoofIp(pwaconf))
	return;

    if (IsConfSpoofNotice(pwaconf))
	sendto_ops("%^C has masked their hostname.", sptr);

    sendto_one_person(sptr, NULL, TOK1_NOTICE, "AUTH :*** Your hostname has been masked.");

    sptr->user->real_oper_host = MyMalloc(strlen(sptr->user->host) + 1);
    sptr->user->real_oper_ip = MyMalloc(strlen(sptr->hostip) + 1);

    strcpy(sptr->user->real_oper_host, sptr->user->host);
    strcpy(sptr->user->real_oper_ip, sptr->hostip);

    if (ServerOpts.opers_can_choose_fakehost) {
	if (sptr->fakehost_override[0]) {
	    strlcpy_irc(sptr->sockhost, sptr->fakehost_override, HOSTLEN);
	    strlcpy_irc(sptr->user->host, sptr->fakehost_override, HOSTLEN);
	} else if (pwaconf->localhost[0]) {
	    strlcpy_irc(sptr->sockhost, pwaconf->localhost, HOSTLEN);
	    strlcpy_irc(sptr->user->host, pwaconf->localhost, HOSTLEN);
	} else {
	    strlcpy_irc(sptr->sockhost, ServerOpts.staffhost, HOSTLEN);
	    strlcpy_irc(sptr->user->host, ServerOpts.staffhost, HOSTLEN);
	}
    } else if (pwaconf != NULL) {
	if (pwaconf->localhost[0]) {
	    strlcpy_irc(sptr->sockhost, pwaconf->localhost, HOSTLEN);
	    strlcpy_irc(sptr->user->host, pwaconf->localhost, HOSTLEN);
	} else {
	    strlcpy_irc(sptr->sockhost, ServerOpts.staffhost, HOSTLEN);
	    strlcpy_irc(sptr->user->host, ServerOpts.staffhost, HOSTLEN);
	}
    }

    throttle_remove(sptr->hostip);

    sptr->flags |= FLAGS_GOTID;
    IN4_ADDR(sptr->ip) = 0;
    strcpy(sptr->hostip, "0.0.0.0");

    return;
}

/* 
 * user_welcome
 *
 * inputs       - client pointer to client to welcome
 * output       - NONE
 * side effects -
 */
static void user_welcome(struct Client *sptr, struct ConfItem *aconf)
{
    aMotd *smotd;
    aMotdItem *smotdline;

    if (ServerOpts.wingate_notice) {
	sendto_one_person(sptr, NULL, TOK1_NOTICE,
			  "AUTH :*** This server runs an open proxy/wingate detection monitor from %s",
			  ServerOpts.monitor_host);
	sendto_one_person(sptr, NULL, TOK1_NOTICE,
			  "AUTH :*** If you do not accept this, disconnect now. "
			  "Further information is available at %s", ServerOpts.proxy_url);
    }

    send_me_numeric(sptr, RPL_WELCOME, ServerInfo.networkname, sptr->name,
		    sptr->user->username, sptr->user->host);
    /* We require the above line for DCC to work, otherwise,   
     * we could use %C instead -TimeMr14C */
    send_me_numeric(sptr, RPL_YOURHOST, get_client_name(&me, TRUE), version);
    send_me_numeric(sptr, RPL_CREATED, creation);
    send_me_numeric(sptr, RPL_MYINFO, me.name, version);
    send_supported(sptr);

    manage_faked_oper_hostname(sptr, aconf);

    send_lusers(sptr, sptr, 1, NULL);

    send_me_notice(sptr, ":*** Notice -- motd was last changed at %s", (&(GeneralOpts.motd))->lastchange);
    if (ServerOpts.short_motd) {
	send_me_notice(sptr, ":*** Notice -- Please read the motd if you haven't read it");
	send_me_numeric(sptr, RPL_MOTDSTART, me.name);
	if ((smotd = &(GeneralOpts.shortmotd)) == NULL) {
	    send_me_numeric(sptr, RPL_MOTD, "*** This is the short motd ***");
	} else {
	    smotdline = smotd->content;
	    while (smotdline) {
		send_me_numeric(sptr, RPL_MOTD, smotdline->line);
		smotdline = smotdline->next;
	    }
	}
	send_me_numeric(sptr, RPL_ENDOFMOTD);
    } else {
	(void) send_message_file(&(GeneralOpts.motd), sptr, sptr, 1, NULL);
    }
}

/*
 * is_silenced - Returns 1 if a sptr is silenced by acptr
 */
int is_silenced(aClient *sptr, aClient *acptr)
{
    dlink_node *lp;
    anUser *user;
    char *cp;
    char sender[HOSTLEN + NICKLEN + USERLEN + 5];

    assert(NULL != acptr);
    assert(NULL != sptr);

    if (!(acptr->user) || !(acptr->user->silence.head) || !(sptr->user))
	return 0;
    user = sptr->user;
    ircsprintf(sender, "%s!%s@%s", sptr->name, user->username, user->host);
    for (lp = acptr->user->silence.head; lp; lp = lp->next) {
	cp = lp->data;
	if (!match(cp, sender)) {
	    if (!MyConnect(sptr))
		sendto_one_server(sptr->from, acptr, TOK1_SILENCE, "%~C :%s", sptr, cp);
	    return 1;
	}
    }
    return 0;
}

int del_silence(aClient *sptr, char *mask)
{
    dlink_node **lp;
    char *cp = NULL;

    for (lp = &(sptr->user->silence.head); *lp; lp = &((*lp)->next)) {
	cp = (*lp)->data;
	if ((cp != NULL) && (irc_strcmp(mask, cp) == 0)) {
	    dlinkDeleteNode(*lp, &sptr->user->silence);
	    return 0;
	}
    }
    return 1;
}

int add_silence(aClient *sptr, char *mask)
{
    dlink_node *lp;
    char *cp;
    int cnt = 0, len = 0;

    for (lp = sptr->user->silence.head; lp; lp = lp->next) {
	cp = lp->data;
	len += strlen(cp);
	if (MyClient(sptr)) {
	    if ((len > MAXSILELENGTH) || (++cnt >= MAXSILES)) {
		send_me_numeric(sptr, ERR_SILELISTFULL, mask);
		return -1;
	    } else {
		if (!match(cp, mask))
		    return -1;
	    }
	} else if (!irc_strcmp(cp, mask))
	    return -1;
    }
    lp = make_dlink_node();
    DupString(cp, mask);
    lp->data = cp;
    dlinkAdd(cp, lp, &sptr->user->silence);
    return 0;
}

int allow_dcc(aClient *to, aClient *from)
{
    dlink_node *lp;
    plist *dl;

    for (lp = to->user->dccallow.head; lp; lp = lp->next) {
	dl = lp->data;
	if (dl->flags == DCC_LINK_ME && dl->cptr == from)
	    return 1;
    }
    return 0;
}

void send_umode(aClient *cptr, aClient *sptr, int old, int sendmask, char *umode_buf)
{

    int *s, flag;
    char *m;
    int what = MODE_NULL;

    /*
     * build a string in umode_buf to represent the change in the user's
     * mode between the new (sptr->flag) and 'old'.   
     */

    m = umode_buf;
    *m = '\0';

    for (s = user_modes; (flag = *s); s += 2) {
	if (MyClient(sptr) && !(flag & sendmask))
	    continue;
	if ((flag & old) && !(sptr->umode & flag)) {
	    if (what == CMODE_DEL)
		*m++ = *(s + 1);
	    else {
		what = CMODE_DEL;
		*m++ = '-';
		*m++ = *(s + 1);
	    }
	} else if (!(flag & old) && (sptr->umode & flag)) {
	    if (what == CMODE_ADD)
		*m++ = *(s + 1);
	    else {
		what = CMODE_ADD;
		*m++ = '+';
		*m++ = *(s + 1);
	    }
	}
    }
    *m = '\0';
    if (*umode_buf && cptr)
	sendto_one_person(cptr, sptr, TOK1_MODE, "%~C :%s", sptr, umode_buf);
}

    /*
     * added Sat Jul 25 07:30:42 EST 1992
     * extra argument evenTS added to send to TS servers or not -orabidoo
     * extra argument evenTS no longer needed with TS only th+hybrid server
     * -Dianora
     */

void send_umode_out(aClient *cptr, aClient *sptr, int old)
{
    aClient *acptr;
    dlink_node *ptr;
    send_umode(NULL, sptr, old, SEND_UMODES, buf);
    for (ptr = serv_list.head; ptr; ptr = ptr->next) {
	acptr = ptr->data;
	if (!acptr)
	    continue;
	if ((acptr != cptr) && (acptr != sptr) && (*buf)) {
	    sendto_one_server(acptr, sptr, TOK1_MODE, "%~C :%s", sptr, buf);
	}
    }
    sendto_service(SERVICE_SEE_UMODES, 0, sptr, NULL, TOK1_MODE, "%s :%s", sptr->name, buf);
    if (cptr && MyClient(cptr))
	send_umode(cptr, sptr, old, SEND_UMODES, buf);
}
