/*
 * IRC - Internet Relay Chat, src/class.c Copyright (C) 1990 Darren
 * Reed
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation,
 * Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: class.c,v 5.1 2002/06/29 12:20:46 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "numeric.h"
#include "sys.h"
#include "h.h"

#define BAD_CONF_CLASS		-1
#define BAD_PING		-2
#define BAD_CLIENT_CLASS	-3

static int log_noclass = -1;
static int log_noping = -1;
static int log_return = -1;

aClass *classes;

int get_conf_class(aConfItem *aconf)
{
    if ((aconf) && Class(aconf))
	return (ConfClass(aconf));

    logevent_call(log_noclass, (aconf) ? aconf->name : "*No Conf*");

    return (BAD_CONF_CLASS);

}

static int get_conf_ping(aConfItem *aconf)
{
    if ((aconf) && Class(aconf))
	return (ConfPingFreq(aconf));

    logevent_call(log_noping, (aconf) ? aconf->name : "*No Conf*");

    return (BAD_PING);
}

int get_client_class(aClient *acptr)
{
    dlink_node *tmp;
    aClass *cl;
    aConfItem *aconf;
    int retc = BAD_CLIENT_CLASS;

    if (acptr && !IsMe(acptr) && (acptr->confs.head))
	for (tmp = acptr->confs.head; tmp; tmp = tmp->next) {
	    aconf = tmp->data;
	    if (!aconf || !(cl = aconf->class))
		continue;
	    if (Class(cl) > retc)
		retc = Class(cl);
	}

    logevent_call(log_return, retc, acptr->name);
    return (retc);
}

int get_client_ping(aClient *acptr)
{
    int ping = 0, ping2;
    aConfItem *aconf;
    dlink_node *dlink;

    dlink = acptr->confs.head;

    if (dlink)
	while (dlink) {
	    aconf = dlink->data;
	    if (aconf->status & (CONF_CLIENT | CONF_SERVER)) {
		ping2 = get_conf_ping(aconf);
		if ((ping2 != BAD_PING) && ((ping > ping2) || !ping))
		    ping = ping2;
	    }
	    dlink = dlink->next;
    } else {
	ping = PINGFREQUENCY;
    }
    if (ping <= 0)
	ping = PINGFREQUENCY;
    return (ping);
}

int get_con_freq(aClass * clptr)
{
    if (clptr)
	return (ConFreq(clptr));
    else
	return (CONNECTFREQUENCY);
}

/*
 * When adding a class, check to see if it is already present first. if
 * so, then update the information for that class, rather than create a
 * new entry for it and later delete the old entry. if no present entry
 * is found, then create a new one and add it in immediately after the
 * first one (class 0).
 */
void add_class(int class, int ping, int confreq, int maxli, long sendq)
{
    aClass *t, *p;

    t = find_class(class);
    if ((t == classes) && (class != 0)) {
	p = (aClass *) make_class();
	NextClass(p) = NextClass(t);
	NextClass(t) = p;
    } else
	p = t;
    Class(p) = class;

    ConFreq(p) = confreq;
    PingFreq(p) = ping;
    MaxLinks(p) = maxli;
    MaxSendq(p) = (sendq > 0) ? sendq : MAXSENDQLENGTH;
    if (p != t)
	Links(p) = 0;
}

aClass *find_class(int cclass)
{
    aClass *cltmp;

    for (cltmp = FirstClass(); cltmp; cltmp = NextClass(cltmp))
	if (Class(cltmp) == cclass)
	    return cltmp;
    return classes;
}

void check_class()
{
aClass *cltmp, *cltmp2;

    for (cltmp2 = cltmp = FirstClass(); cltmp; cltmp = NextClass(cltmp2)) {
	if (MaxLinks(cltmp) < 0) {
	    NextClass(cltmp2) = NextClass(cltmp);
	    if (Links(cltmp) <= 0)
		free_class(cltmp);
	} else
	    cltmp2 = cltmp;
    }
}

void initclass()
{
    classes = (aClass *) make_class();

    Class(FirstClass()) = 0;

    ConFreq(FirstClass()) = CONNECTFREQUENCY;
    PingFreq(FirstClass()) = PINGFREQUENCY;
    MaxLinks(FirstClass()) = MAXIMUM_LINKS;
    MaxSendq(FirstClass()) = MAXSENDQLENGTH;
    Links(FirstClass()) = 0;
    NextClass(FirstClass()) = NULL;

    log_noclass = logevent_register("no class", LOG_ON_LOG, LOG_IRCLOG, LOG_NOTICE, 
				    "No Class For %s");
    log_noping = logevent_register("no ping", LOG_ON_LOG, LOG_IRCLOG, LOG_NOTICE, 
                                    "No Ping For %s");
    log_return = logevent_register("return class", LOG_ON_LOG, LOG_IRCLOG, LOG_NOTICE, 
                                   "Returning Class %d For %s");
}

void report_classes(aClient *sptr)
{
    aClass *cltmp;

    for (cltmp = FirstClass(); cltmp; cltmp = NextClass(cltmp))
	send_me_numeric(sptr, RPL_STATSYLINE,
			'Y', Class(cltmp), PingFreq(cltmp), ConFreq(cltmp),
			MaxLinks(cltmp), MaxSendq(cltmp));
}

long get_sendq(aClient *cptr)
{
    int sendq = MAXSENDQLENGTH, retc = BAD_CLIENT_CLASS;
    dlink_node *tmp;
    aClass *cl;
    aConfItem *aconf;

    if (cptr && !IsMe(cptr) && (cptr->confs.head))
	for (tmp = cptr->confs.head; tmp; tmp = tmp->next) {
	    aconf = tmp->data;
	    if (!aconf || !(cl = aconf->class))
		continue;
	    if (Class(cl) > retc) {
		retc = Class(cl);
		sendq = MaxSendq(cl);
	    }
	}
    return sendq;
}

aClass *make_class()
{
aClass *tmp;

    tmp = (aClass *) MyMalloc(sizeof(aClass));
    return tmp;
}

void free_class(tmp)
     aClass *tmp;
{
    MyFree((char *) tmp);
}
