/*
 *   Copyright (C) 2000-2002 TR-IRCD Development
 *
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Co Center
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dynlink.c,v 5.7 2002/08/11 11:17:00 tr-ircd Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>

#include <limits.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "confitem.h"
#include "h.h"
#include "hook.h"
#include "s_conf.h"
#include "msg.h"
#include "numeric.h"
#include "modules.h"

#ifndef STATIC_MODULES

static int hookid_load_module = 0;
static int hookid_unload_module = 0;
static int undef_symbol = -1;
static int toomany_symbols = -1;
static int link_editor = -1;
static int module_error = -1;

void init_loader(void *unused)
{ 
    hookid_load_module = hook_add_event("load module");
    hookid_unload_module = hook_add_event("unload module");

    undef_symbol = logevent_register("undefined symbol", LOG_ON_LOG, LOG_IRCLOG | LOG_ERRORLOG, LOG_ERROR,
			  "Undefined symbol: %s");
    toomany_symbols = logevent_register("too many symbols", LOG_ON_LOG, LOG_IRCLOG | LOG_ERRORLOG, LOG_ERROR,
			  "Symbol `%s' found in `%s' and `%s'");
    link_editor = logevent_register("link editor error", LOG_ON_LOG, LOG_IRCLOG | LOG_ERRORLOG, LOG_ERROR,
			  "Link editor error: %s for %s");
    module_error = logevent_register("module load error", LOG_ON_LOG, LOG_IRCLOG | LOG_ERRORLOG, LOG_ERROR,
			  "Error loading module %s: %s");
}

#ifndef RTLD_NOW
#define RTLD_NOW RTLD_LAZY	/* openbsd deficiency */
#endif

#ifdef HAVE_MACH_O_DYLD_H
/*
 * ** jmallett's dl*(3) shims for NSModule(3) systems.
 */
#include <mach-o/dyld.h>

#ifndef HAVE_DLOPEN
#ifndef RTLD_LAZY
#define RTLD_LAZY 2185		/* built-in dl*(3) don't care */
#endif

void undefinedErrorHandler(const char *);
NSModule multipleErrorHandler(NSSymbol, NSModule, NSModule);
void linkEditErrorHandler(NSLinkEditErrors, int, const char *, const char *);
char *dlerror(void);
void *dlopen(char *, int);
int dlclose(void *);
void *dlsym(void *, char *);

static int firstLoad = TRUE;
static int myDlError;
static char *myErrorTable[] = { "Loading file as object failed\n",
    "Loading file as object succeeded\n",
    "Not a valid shared object\n",
    "Architecture of object invalid on this architecture\n",
    "Invalid or corrupt image\n",
    "Could not access object\n",
    "NSCreateObjectFileImageFromFile failed\n",
    NULL
};

void undefinedErrorHandler(const char *symbolName)
{
    sendto_ops("Undefined symbol: %s", symbolName);
    logevent_call(undef_symbol, symbolName);
    return;
}

NSModule multipleErrorHandler(NSSymbol s, NSModule old, NSModule new)
{
    /* XXX
     * ** This results in substantial leaking of memory... Should free one
     * ** module, maybe?
     */
    sendto_ops("Symbol `%s' found in `%s' and `%s'",
	       NSNameOfSymbol(s), NSNameOfModule(old), NSNameOfModule(new));
    logevent_call(toomany_symbols, NSNameOfSymbol(s), NSNameOfModule(old), NSNameOfModule(new));
    /* We return which module should be considered valid, I believe */
    return new;
}

void linkEditErrorHandler(NSLinkEditErrors errorClass, int errnum,
			  const char *fileName, const char *errorString)
{
    sendto_ops("Link editor error: %s for %s", errorString, fileName);
    logevent_call(link_editor, errorString, fileName);
    return;
}

char *dlerror(void)
{
    return myDlError == NSObjectFileImageSuccess ? NULL : myErrorTable[myDlError % 7];
}

void *dlopen(char *filename, int unused)
{
    NSObjectFileImage myImage;
    NSModule myModule;

    if (firstLoad) {
	/*
	 * ** If we are loading our first symbol (huzzah!) we should go ahead
	 * ** and install link editor error handling!
	 */
    NSLinkEditErrorHandlers linkEditorErrorHandlers;

	linkEditorErrorHandlers.undefined = undefinedErrorHandler;
	linkEditorErrorHandlers.multiple = multipleErrorHandler;
	linkEditorErrorHandlers.linkEdit = linkEditErrorHandler;
	NSInstallLinkEditErrorHandlers(&linkEditorErrorHandlers);
	firstLoad = FALSE;
    }
    myDlError = NSCreateObjectFileImageFromFile(filename, &myImage);
    if (myDlError != NSObjectFileImageSuccess) {
	return NULL;
    }
    myModule = NSLinkModule(myImage, filename, NSLINKMODULE_OPTION_PRIVATE);
    return (void *) myModule;
}

int dlclose(void *myModule)
{
    NSUnLinkModule(myModule, FALSE);
    return 0;
}

void *dlsym(void *myModule, char *mySymbolName)
{
    NSSymbol mySymbol;

    mySymbol = NSLookupSymbolInModule((NSModule) myModule, mySymbolName);
    return NSAddressOfSymbol(mySymbol);
}
#endif 	/* HAVE_DLOPEN */
#endif	/* MACH_O_DYLD_H */

#define MODS_INCREMENT 10

static char unknown_ver[] = "<unknown>";

/*
 * unload_one_module 
 *
 * inputs       - name of module to unload
 *              - 1 to say modules unloaded, 0 to not
 * output       - 0 if successful, -1 if error
 * side effects - module is unloaded
 */
int unload_one_module(char *name, int check)
{
    int modindex;
    void (*deinitfunc) (void) = NULL;
    struct hook_data thisdata;

    if ((modindex = findmodule_byname(name)) == -1)
	return -1;

    thisdata.name = name;
    if (hook_call_event(hookid_unload_module, &thisdata)) {
	if (check == 1)
	    sendto_ops("Module %s has been denied usage.", name);
	return -2;
    }

    if (modlist[modindex]->persistent) {
	if (check == 1)
	    sendto_ops("Module %s cannot be unloaded.", name);
  	return -2;
    }

#if defined(HAVE_SHL_LOAD)
    if (shl_findsym(modlist[modindex]->address, "_moddeinit", TYPE_UNDEFINED, &deinitfunc) == 0)
	deinitfunc();
    else if (shl_findsym
	     (modlist[modindex]->address, "__moddeinit", TYPE_UNDEFINED, &deinitfunc) == 0)
	deinitfunc();
    shl_unload((shl_t) & (modlist[modindex]->address));
#else
    if ((deinitfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "_moddeinit"))
	|| (deinitfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "__moddeinit"))) {
	deinitfunc();
    }
    dlclose(modlist[modindex]->address);
#endif

    logevent_call(LogSys.modunload, name);

    MyFree(modlist[modindex]->name);
    memcpy(&modlist[modindex], &modlist[modindex + 1],
	   sizeof(struct module) * ((num_mods - 1) - modindex));

    if (num_mods != 0)
	num_mods--;

    if (check == 1) 
	sendto_ops("Module %s unloaded", name);

    return 0;
}

/*
 * load_a_module
 *
 * inputs       - path name of module
 * output       - -1 if error 0 if success
 * side effects - loads a module if successful
 */
int load_a_module(char *path, int check)
{
#if defined(HAVE_SHL_LOAD)
    shl_t tmpptr;
#else
    void *tmpptr = NULL;
#endif
    char *mod_basename;
    void (*initfunc) (void) = NULL;
    char **verp;
    char *ver;
    struct module **new_modlist = NULL;
    struct hook_data thisdata;
    int *dependent = 0;
    int depend = 0;
    int *persistent = 0;
    int persist = 0;

    mod_basename = irc_basename(path);

    thisdata.name = mod_basename;
    if (hook_call_event(hookid_load_module, &thisdata))
        return -1;

#if defined(HAVE_SHL_LOAD)
    tmpptr = shl_load(path, BIND_IMMEDIATE, NULL);
#else
    tmpptr = dlopen(path, RTLD_NOW);
#endif

    if (tmpptr == NULL) {

#if defined(HAVE_SHL_LOAD)
    const char *err = strerror(errno);
#else
    const char *err = dlerror();
#endif

	logevent_call(module_error, mod_basename, err);
	sendto_ops("Error loading module %s: %s", mod_basename, err);
	MyFree(mod_basename);
	return -1;
    }
#if defined(HAVE_SHL_LOAD)
    if (shl_findsym(&tmpptr, "_modinit", TYPE_UNDEFINED, (void *) &initfunc) == -1) {
	if (shl_findsym(&tmpptr, "__modinit", TYPE_UNDEFINED, (void *) &initfunc) == -1) {
	    logevent_call(LogSys.modnoinit, mod_basename);
	    sendto_ops("Module %s has no _modinit() function", mod_basename);
	    shl_unload(tmpptr);
	    MyFree(mod_basename);
	    return -1;
	}
    }

    if (shl_findsym(&tmpptr, "_version", TYPE_UNDEFINED, &verp) == -1) {
	if (shl_findsym(&tmpptr, "__version", TYPE_UNDEFINED, &verp) == -1)
	    ver = unknown_ver;
	else
	    ver = *verp;
    } else
	ver = *verp;

    if (shl_findsym(&tmpptr, "_dependent", TYPE_UNDEFINED, &dependent) == -1) {
        if (shl_findsym(&tmpptr, "__dependent", TYPE_UNDEFINED, &dependent) != -1)
            depend = *dependent;
    }

    if (shl_findsym(&tmpptr, "_persistent", TYPE_UNDEFINED, &persistent) == -1) {
        if (shl_findsym(&tmpptr, "__persistent", TYPE_UNDEFINED, &persistent) != -1)
            persist = *persistent;
    } 

#else
    initfunc = (void (*)(void)) (uintptr_t) dlsym(tmpptr, "_modinit");
    if (initfunc == NULL && (initfunc = (void (*)(void)) (uintptr_t) dlsym(tmpptr, "__modinit")) == NULL) {
	logevent_call(LogSys.modnoinit, mod_basename);
	sendto_ops("Module %s has no _modinit() function", mod_basename);
	(void) dlclose(tmpptr);
	MyFree(mod_basename);
	return -1;
    }

    verp = (char **) dlsym(tmpptr, "_version");
    if (verp == NULL && (verp = (char **) dlsym(tmpptr, "__version")) == NULL)
	ver = unknown_ver;
    else
	ver = *verp;

    dependent = (int *) dlsym(tmpptr, "_dependent");
    if (dependent || (dependent = (int *) dlsym(tmpptr, "__dependent")))
	depend = *dependent;

    persistent = (int *) dlsym(tmpptr, "_persistent");
    if (persistent || (persistent = (int *) dlsym(tmpptr, "__persistent")))
        persist = *persistent;

#endif

    /* moved increase_modlist() in here -TimeMr14C */

    if ((num_mods + 1) >= max_mods) {
	new_modlist = (struct module **) MyMalloc(sizeof(struct module) *
						  (max_mods + MODS_INCREMENT));
	memcpy((void *) new_modlist, (void *) modlist, sizeof(struct module) * num_mods);
	MyFree(modlist);
	modlist = new_modlist;
	max_mods += MODS_INCREMENT;
    }

    modlist[num_mods] = (struct module *) MyMalloc(sizeof(struct module));
    modlist[num_mods]->address = tmpptr;
    modlist[num_mods]->version = ver;
    modlist[num_mods]->name = strdup(mod_basename);
    modlist[num_mods]->dependent = depend;
    modlist[num_mods]->persistent = persist;
    num_mods++;
    initfunc();
    if (check == 1) {
	sendto_ops("Module %s [version: %s] loaded at 0x%lx",
		   mod_basename, ver, (unsigned long) tmpptr);
    }

    logevent_call(LogSys.modload, mod_basename);

    MyFree(mod_basename);
    return 0;
}

int deinit_reinit_module(char *name)
{
    int modindex;
    void (*deinitfunc) (void) = NULL;
    void (*initfunc) (void) = NULL;
    struct hook_data thisdata;

    if ((modindex = findmodule_byname(name)) == -1)
	return -1;

    thisdata.name = name;
    if (hook_call_event(hookid_unload_module, &thisdata)) 
	return -2;

    if (modlist[modindex]->persistent) 
  	return -2;

#if defined(HAVE_SHL_LOAD)
    if (shl_findsym(modlist[modindex]->address, "_moddeinit", TYPE_UNDEFINED, &deinitfunc) == 0)
	deinitfunc();
    else if (shl_findsym
	     (modlist[modindex]->address, "__moddeinit", TYPE_UNDEFINED, &deinitfunc) == 0)
	deinitfunc();
#else
    if ((deinitfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "_moddeinit"))
	|| (deinitfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "__moddeinit"))) {
	deinitfunc();
    }
#endif

#if defined(HAVE_SHL_LOAD)
    if (shl_findsym(modlist[modindex]->address, "_modinit", TYPE_UNDEFINED, &initfunc) == 0)
	initfunc();
    else if (shl_findsym
	     (modlist[modindex]->address, "__modinit", TYPE_UNDEFINED, &initfunc) == 0)
	initfunc();
#else
    if ((initfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "_modinit"))
	|| (initfunc = (void (*)(void)) (uintptr_t) dlsym(modlist[modindex]->address, "__modinit"))) {
	initfunc();
    }
#endif

    logevent_call(LogSys.modunload, name);
    logevent_call(LogSys.modload, name);

    return 0;
}

#else

void init_loader(void *unused)
{
    return; 
}

#endif /* STATIC_MODULES */
