/************************************************************************
 *   IRC - Internet Relay Chat, src/hash.c
 *   Copyright (C) 1991 Darren Reed
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * $Id: hash.c,v 5.9 2003/05/11 19:33:17 tr-ircd Exp $ 
 */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "hash.h"
#include "s_conf.h"
#include "h.h"

static aHashEntry clientTable[U_MAX];
static aHashEntry channelTable[CH_MAX];

/*
 * look in whowas.c for the missing ...[WW_MAX]; entry - Dianora */
/*
 * Hashing.
 * 
 * The server uses a chained hash table to provide quick and efficient
 * hash table mantainence (providing the hash function works evenly
 * over the input range).  The hash table is thus not susceptible to
 * problems of filling all the buckets or the need to rehash. It is
 * expected that the hash table would look somehting like this during
 * use: +-----+    +-----+    +-----+   +-----+ 
 *   ---| 224 |----| 225 |----| 226 |---| 227 |--- 
 *      +-----+    +-----+    +-----+   +-----+ 
 *         |          |          | 
 *      +-----+    +-----+    +-----+ 
 *      |  A  |    |  C  |    |  D  | 
 *      +-----+    +-----+    +-----+ 
 *         | 
 *      +-----+ 
 *      |  B  | 
 *      +-----+
 * 
 * A - GOPbot, B - chang, C - hanuaway, D - *.mu.OZ.AU
 * 
 * The order shown above is just one instant of the server.  Each time a
 * lookup is made on an entry in the hash table and it is found, the
 * entry is moved to the top of the chain.
 * 
 * ^^^^^^^^^^^^^^^^ **** Not anymore - Dianora
 * 
 */

static unsigned
int hash_nick_name(const char *name)
{
    unsigned int h = 0;

    while (*name) {
	h = (h << 4) - (h + (unsigned char) ToLower(*name++));
    }

    return (h & (U_MAX - 1));
}

/*
 * hash_channel_name
 *
 * calculate a hash value on at most the first 30 characters of the channel
 * name. Most names are short than this or dissimilar in this range. There
 * is little or no point hashing on a full channel name which maybe 255 chars
 * long.
 */
static unsigned
int hash_channel_name(const char *name)
{
    register int i = 30;
    unsigned int h = 0;

    while (*name && --i) {
	h = (h << 4) - (h + (unsigned char) ToLower(*name++));
    }

    return (h & (CH_MAX - 1));
}

unsigned int hash_whowas_name(char* name)
{
  unsigned int h = 0;

  while (*name)
    {
      h = (h << 4) - (h + (unsigned char)ToLower(*name++));
    }
  return(h & (WW_MAX - 1));
}


void clear_client_hash_table()
{
    memset((char *) clientTable, '\0', sizeof(aHashEntry) * U_MAX);
}

void clear_channel_hash_table()
{
    memset((char *) channelTable, '\0', sizeof(aHashEntry) * CH_MAX);
}

int add_to_client_hash_table(char *name, aClient *cptr)
{
    unsigned int hashv;

    if (name == NULL || cptr == NULL)
        return 0;

    hashv = hash_nick_name(name);
    cptr->hnext = (aClient *) clientTable[hashv].list;
    clientTable[hashv].list = (void *) cptr;
    clientTable[hashv].links++;
    clientTable[hashv].hits++;
    return 0;
}

int add_to_channel_hash_table(char *name, aChannel *chptr)
{
    int hashv;

    if (name == NULL || chptr == NULL)
        return 0;

    hashv = hash_channel_name(name);
    chptr->hnextch = (aChannel *) channelTable[hashv].list;
    channelTable[hashv].list = (void *) chptr;
    channelTable[hashv].links++;
    channelTable[hashv].hits++;
    return 0;
}

int del_from_client_hash_table(char *name, aClient *cptr)
{
    aClient *tmp, *prev = NULL;
    int hashv;

    if (name == NULL || cptr == NULL)
        return 0;

    hashv = hash_nick_name(name);
    for (tmp = (aClient *) clientTable[hashv].list; tmp; tmp = tmp->hnext) {
	if (tmp == cptr) {
	    if (prev)
		prev->hnext = tmp->hnext;
	    else
		clientTable[hashv].list = (void *) tmp->hnext;
	    tmp->hnext = NULL;
	    if (clientTable[hashv].links > 0) {
		--clientTable[hashv].links;
		return 1;
	    } else
		/*
		 * Should never actually return from here and if we do it
		 * is an error/inconsistency in the hash table.
		 */
		return -1;
	}
	prev = tmp;
    }
    return 0;
}

int del_from_channel_hash_table(char *name, aChannel *chptr)
{
    aChannel *tmp, *prev = NULL;
    int hashv;

    if (name == NULL || chptr == NULL)
        return 0;

    hashv = hash_channel_name(name);
    for (tmp = (aChannel *) channelTable[hashv].list; tmp; tmp = tmp->hnextch) {
	if (tmp == chptr) {
	    if (prev)
		prev->hnextch = tmp->hnextch;
	    else
		channelTable[hashv].list = (void *) tmp->hnextch;
	    tmp->hnextch = NULL;
	    if (channelTable[hashv].links > 0) {
		--channelTable[hashv].links;
		return 1;
	    } else
		return -1;
	}
	prev = tmp;
    }
    return 0;
}

aClient *find_client(char *name)
{
    aClient *tmp;
    int hashv;

    hashv = hash_nick_name(name);
    if (hashv <= 0)
	return NULL;
    tmp = (aClient *) clientTable[hashv].list;

    /*
     * Got the bucket, now search the chain.
     */
    for (; tmp; tmp = tmp->hnext) {
	if (!tmp->name)
	    continue;
	if (irc_strcmp(name, tmp->name) == 0) 
	    return (tmp);
    }
    return NULL;
}

aClient *find_server(char *server)
{
    struct Client *tmp;
    unsigned int hashv;

    hashv = hash_nick_name(server);
    if (hashv <= 0)
        return NULL;
    tmp = (struct Client *) clientTable[hashv].list;

    for (; tmp; tmp = tmp->hnext) {
	if (!IsServer(tmp) && !IsMe(tmp))
	    continue;
	if (irc_strcmp(server, tmp->name) == 0) {
	    return tmp;
	}
    }

    return tmp;
}

aChannel *find_channel(char *name)
{
    int hashv;
    aChannel *tmp;

    if (!name)
	return NULL;

    hashv = hash_channel_name(name);
    if (hashv <= 0)
        return NULL;
    tmp = (aChannel *) channelTable[hashv].list;

    for (; tmp; tmp = tmp->hnextch)
	if (irc_strcmp(name, tmp->chname) == 0) 
	    return (tmp);
    return NULL;
}

aChannel *hash_get_chan_bucket(int hashv)
{
    if (hashv > CH_MAX)
	return NULL;
    return (aChannel *) channelTable[hashv].list;
}
